#!/usr/bin/env python3

# Copyright (c) 2018 Unrud <unrud@outlook.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import collections
import contextlib
import functools
import itertools
import json
import os
import subprocess
import tempfile

SCAD_FILE = "CNC-v3_2.scad"
# Data is parsed from the last line the OpenSCAD file prints
SCAD_APPENDIX = """;
function cat(L1, L2) = [for (i=[0:len(L1)+len(L2)-1])
                        i < len(L1)? L1[i] : L2[i-len(L1)]];
echo(cat(SQUARE_TUBES, CROSS_BEAMS));
"""
SCAD_CMD = "openscad"
BASE_FORMAT = {
    "title": "",
    "columns": [
        {
            "title": "Part number",
            "value": lambda e: e[0]
        },
        {
            "title": "Length (mm)",
            "value": lambda e: fmt_unit(e[1])
        }
    ],
    "filter": lambda e: False,
    "name": lambda e: e[0],
    "length": lambda e: e[1],
    "drillholes": lambda e: e[2]
}
FORMATS = [
    {
        **BASE_FORMAT,
        "title": "Framework",
        "filter": lambda e: e[0].startswith("F")
    },
    {
        **BASE_FORMAT,
        "title": "Cross beams framework",
        "columns": BASE_FORMAT["columns"] + [
            {
                "title": "Hole distance (mm)",
                "value": lambda e: fmt_unit(abs(e[2][-1][0] - e[2][0][0]))
            },
            {
                "title": "Center hole diameter (mm)",
                "value": lambda e: ("-" if len(e[2]) == 2
                                    else fmt_unit(e[2][1][1]))
            }
        ],
        "filter": lambda e: e[0].startswith("C"),
        "drillholes": lambda e: []
    },
    {
        **BASE_FORMAT,
        "title": "X Axis",
        "filter": lambda e: e[0].startswith("X")
    },
    {
        **BASE_FORMAT,
        "title": "Table beams",
        "filter": lambda e: e[0].startswith("T"),
        "drillholes": lambda e: [e[2]]
    },
    {
        **BASE_FORMAT,
        "title": "Y Axis",
        "filter": lambda e: e[0].startswith("Y")
    },
    {
        **BASE_FORMAT,
        "title": "Z Axis",
        "filter": lambda e: e[0].startswith("Z")
    }
]


def fmt_unit(f):
    s = "%.2f" % float(f)
    return s.rstrip("0").rstrip(".") if "." in s else s


def normalize_drillholes(drillholes, length):
    drillholes = tuple(tuple(sorted((float(e[0]), float(e[1])) for e in s))
                       for s in drillholes)
    rotated_drillholes = [drillholes[i:] + drillholes[:i]
                          for i in range(len(drillholes))]
    reversed_rotated_drillholes = [tuple(tuple(sorted((length - e[0], e[1])
                                                      for e in s)) for s in xs)
                                   for xs in rotated_drillholes]
    # Sort empty tuples to the end
    cmp = lambda a, b: 0 if a == b else (1 if a > b and b or not a else -1)
    return max(rotated_drillholes + reversed_rotated_drillholes,
               key=lambda x: tuple(reversed(list(map(functools.cmp_to_key(cmp),
                                                     x)))))


Key = collections.namedtuple("Key", ["format_id", "length", "norm_drillholes"])


def make_key(_format, entry):
    length = float(_format["length"](entry))
    drillholes = normalize_drillholes(_format["drillholes"](entry), length)
    return Key(id(_format), length, drillholes)


def main():
    with contextlib.ExitStack() as stack:
        with open(SCAD_FILE) as f, tempfile.NamedTemporaryFile(
                "w", prefix=".tmp-", suffix=os.path.splitext(SCAD_FILE)[1],
                dir=os.path.dirname(SCAD_FILE), delete=False) as temp_scad:
            stack.callback(os.remove, temp_scad.name)
            temp_scad.write(f.read())
            temp_scad.write(SCAD_APPENDIX)
        with tempfile.NamedTemporaryFile(
                suffix=".echo", delete=False) as temp_echo:
            stack.callback(os.remove, temp_echo.name)
        subprocess.run([SCAD_CMD, "-o", temp_echo.name, temp_scad.name])
        with open(temp_echo.name) as f:
            echo_lines = [l[len("ECHO: "):] for l in f.readlines()
                          if l.startswith("ECHO: ")]
    entries = json.loads(echo_lines[-1])

    print("# Dimensions")
    drillhole_buckets = collections.OrderedDict()
    for f in FORMATS:
        print("\n## %s\n" % f["title"])
        for c in f["columns"]:
            print("| %s " % c["title"], end="")
        print("|")
        for c in f["columns"]:
            print("|-", end="")
        print("|")
        for e in filter(f["filter"], entries):
            for c in f["columns"]:
                print("| %s " % c["value"](e), end="")
            print("|")
            if not f["drillholes"](e):
                continue
            key = make_key(f, e)
            if key not in drillhole_buckets:
                drillhole_buckets[key] = []
            drillhole_buckets[key].append(f["name"](e))

    print("\n# Drill holes")
    for key, names in drillhole_buckets.items():
        title = " + ".join(names)
        print("\n## %s (%smm)\n" % (title, fmt_unit(key.length)))
        slots = sorted(
            {e[0] for e in itertools.chain.from_iterable(key.norm_drillholes)})
        for _ in slots:
            print("| ", end="")
        print("|")
        for _ in slots:
            print("|-", end="")
        print("|")
        for xs in key.norm_drillholes:
            for p in slots:
                print("| ", end="")
                for s in xs:
                    if s[0] == p:
                        print("%s (%s)" % (fmt_unit(s[0]), fmt_unit(s[1])),
                              end="")
                        break
                else:
                    print("-", end="")
            print(" |")


if __name__ == "__main__":
    main()